from flask_appbuilder.security.views import UserDBModelView
from flask_appbuilder import ModelView
from .s3_sec_models import S3Access, S3Group, S3Endpoint
from flask_appbuilder.models.sqla.interface import SQLAInterface

from flask_babel import lazy_gettext


class S3EndpointModelView(ModelView):
    route_base = '/s3endpoints'
    datamodel = SQLAInterface(S3Endpoint)
    list_columns = [
        'name',
        'url',
        's3_accesses'
    ]
    edit_columns = [
        'name',
        'url',
        's3_accesses',
        's3_default_region',
        'trust_ca_bundle'
    ]
    add_columns = [
        'name',
        'url',
        's3_accesses',
        's3_default_region',
        'trust_ca_bundle'
    ]


class S3AccessModelView(ModelView):
    route_base = '/s3access'
    datamodel = SQLAInterface(S3Access)
    list_columns = [
        'name',
        's3_access_key',
        's3_endpoint',
        'users',
        'groups'
    ]
    edit_columns = [
        'name',
        's3_access_key',
        's3_secret_key',
        's3_endpoint',
        'users',
        'groups'
    ]
    add_columns = [
        'name',
        's3_access_key',
        's3_secret_key',
        's3_endpoint',
        'users',
        'groups'
    ]


class S3GroupModelView(ModelView):
    route_base = '/s3group'
    datamodel = SQLAInterface(S3Group)
    list_columns = [
        'name',
        'users',
        's3_accesses'
    ]
    edit_columns = [
        'name',
        'users',
        's3_accesses'
    ]
    add_columns = [
        'name',
        'users',
        's3_accesses'
    ]


class S3UserDBModelView(UserDBModelView):
    """
        View that add DB specifics to User view.
        Override to implement your own custom view.
        Then override userdbmodelview property on SecurityManager
    """
    related_views = [S3GroupModelView, S3AccessModelView]

    show_fieldsets = [
        (lazy_gettext('User info'),
         {'fields': ['username', 'active', 'roles', 'login_count', 'extra']}),
        (lazy_gettext('Personal Info'),
         {'fields': ['first_name', 'last_name', 'email'], 'expanded': True}),
        (lazy_gettext('Audit Info'),
         {'fields': ['last_login', 'fail_login_count', 'created_on',
                     'created_by', 'changed_on', 'changed_by'], 'expanded': False}),
    ]
    user_show_fieldsets = [
        (lazy_gettext('User info'),
         {'fields': ['username', 'active', 'roles', 'login_count', 'extra']}),
        (lazy_gettext('Personal Info'),
         {'fields': ['first_name', 'last_name', 'email'], 'expanded': True}),
    ]
    add_columns = [
        'first_name',
        'last_name',
        'username',
        'active',
        'email',
        'roles',
        'password',
        'conf_password',
        's3_accesses',
        'groups'
    ]
    list_columns = [
        'first_name',
        'last_name',
        'username',
        'email',
        'active',
        'roles',
        's3_accesses',
        'groups'
    ]
    edit_columns = [
        'first_name',
        'last_name',
        'username',
        'active',
        'email',
        'roles',
        's3_accesses',
        'groups'
    ]


import html, json

from flask_appbuilder import expose, has_access
from flask_appbuilder.views import BaseView, IndexView
from flask import redirect, session, url_for, request, current_app
from flask_login import current_user


class S3IndexView(IndexView):

    @expose('/')
    def index(self):
        if current_user.is_anonymous:
            return redirect(url_for('AuthDBView.login'))
        else:
            return redirect(url_for('S3View.list'))


class S3View(BaseView):
    route_base = '/s3'

    @expose('/')
    @has_access
    def list(self):
        current_app.logger.debug("S3 Index view called by user %s", current_user.username)
        s3 = session[current_user.id]
        s3.s3WebData.currentSearchPrefix = request.args.get('searchPrefix', default="")
        if s3.s3WebData.currentBucketName != "":
            s3.setBucketItems(s3.s3WebData.currentSearchPrefix)
        else:
            s3.s3WebData.commonPrefixes = []
            s3.s3WebData.contents = []
            s3.s3WebData.pageAmount = 0

        return self.render_template('s3Content.html',
                                    s3WebData=json.dumps(s3.s3WebData, default=lambda o: o.__dict__),
                                    accessList=json.dumps(s3.s3WebData.accessList, default=lambda o: o.__dict__),
                                    currentBucketList=json.dumps(s3.s3WebData.currentBucketList,
                                                                 default=lambda o: o.__dict__),
                                    contents=json.dumps(s3.s3WebData.contents, default=lambda o: o.__dict__),
                                    commonPrefixes=json.dumps(s3.s3WebData.commonPrefixes,
                                                              default=lambda o: o.__dict__),
                                    currentPrefixesList=[]
                                    )

    @expose('/<string:prefixUrl>')
    @has_access
    def get(self, prefixUrl):
        current_app.logger.debug("S3 view with prefix %s called by user %s", prefixUrl, current_user.username)
        s3 = session[current_user.id]
        s3.s3WebData.currentSearchPrefix = request.args.get('searchPrefix', default="")
        prefix = html.unescape(prefixUrl).replace(" ", "/") + s3.s3WebData.currentSearchPrefix
        s3.setBucketItems(prefix)
        content = s3.s3WebData.contents
        if len(content) == 1 and content[0]['Key'] == prefix and content[0]['Size'] >= 0:
            return redirect(s3.getPresignedUrl(prefix), code=302)
        return self.render_template('s3Content.html',
                                    s3WebData=json.dumps(s3.s3WebData, default=lambda o: o.__dict__),
                                    accessList=json.dumps(s3.s3WebData.accessList, default=lambda o: o.__dict__),
                                    currentBucketList=json.dumps(s3.s3WebData.currentBucketList,
                                                                 default=lambda o: o.__dict__),
                                    contents=json.dumps(s3.s3WebData.contents, default=lambda o: o.__dict__),
                                    commonPrefixes=json.dumps(s3.s3WebData.commonPrefixes,
                                                              default=lambda o: o.__dict__),
                                    currentPrefixesList=s3.s3WebData.currentPrefix.split('/')
                                    )




from datetime import date

from flask_appbuilder.security.sqla.models import User
from sqlalchemy import Table, Text, Date
from sqlalchemy.orm import relationship, registry
from flask_appbuilder import Model
from sqlalchemy import Column, Integer, String, ForeignKey

mapper_registry = registry()

s3_assoc_user_s3access = Table('s3_assoc_user_s3access', Model.metadata,
                               Column('user_id', Integer, ForeignKey('ab_user.id'), primary_key=True),
                               Column('s3_access_id', Integer, ForeignKey('s3_access.id'), primary_key=True)
                               )

s3_assoc_group_s3access = Table('s3_assoc_group_s3access', Model.metadata,
                                Column('s3_group_id', Integer, ForeignKey('s3_group.id'), primary_key=True),
                                Column('s3_access_id', Integer, ForeignKey('s3_access.id'), primary_key=True)
                                )

s3_assoc_user_group = Table('s3_assoc_user_group', Model.metadata,
                            Column('user_id', Integer, ForeignKey('ab_user.id'), primary_key=True),
                            Column('s3_group_id', Integer, ForeignKey('s3_group.id'), primary_key=True)
                            )


class S3User(User):
    __tablename__ = 'ab_user'
    groups = relationship('S3Group', secondary=s3_assoc_user_group, back_populates='users')
    s3_accesses = relationship('S3Access', secondary=s3_assoc_user_s3access, back_populates='users')


class S3Endpoint(Model):
    __tablename__ = 's3_endpoint'
    id = Column(Integer, primary_key=True)
    name = Column(String(150), nullable=False, unique=True)
    url = Column(String(150), nullable=False, unique=False)
    trust_ca_bundle = Column(Text, nullable=True, unique=False)
    s3_default_region = Column(String(150), nullable=False, unique=False)
    s3_accesses = relationship("S3Access", back_populates='s3_endpoint')

    def __repr__(self):
        return self.name


class S3Group(Model):
    __tablename__ = 's3_group'
    id = Column(Integer, primary_key=True)
    name = Column(String(150), nullable=False, unique=True)
    users = relationship('S3User', secondary=s3_assoc_user_group, back_populates='groups')
    s3_accesses = relationship('S3Access', secondary=s3_assoc_group_s3access, back_populates='groups')
    begin_date = Column(Date, default=date.today(), nullable=False)
    end_date = Column(Date, nullable=True)

    def __repr__(self):
        return self.name


class S3Access(Model):
    __tablename__ = 's3_access'
    id = Column(Integer, primary_key=True)
    name = Column(String(150), nullable=False, unique=True)
    s3_access_key = Column(String(150), nullable=False, unique=False)
    s3_secret_key = Column(String(150), nullable=False)
    s3_endpoint_id = Column(Integer, ForeignKey('s3_endpoint.id'), nullable=False)
    s3_endpoint = relationship("S3Endpoint", back_populates='s3_accesses')
    users = relationship('S3User', secondary=s3_assoc_user_s3access, back_populates='s3_accesses')
    groups = relationship('S3Group', secondary=s3_assoc_group_s3access, back_populates='s3_accesses')
    begin_date = Column(Date, default=date.today(), nullable=False)
    end_date = Column(Date, nullable=True)

    def __repr__(self):
        return self.name


class Sqls:
    def __init__(self, db):
        self.db = db
        self.session = db.session

    def getEndpointForRegion(self, region):
        return self.db.session.query(S3Endpoint).filter_by(s3_default_region=region).first()

    def setUserAccessView(self):
        print("Info: Create view s3_user_access")
        sql = """CREATE VIEW s3_user_access
        AS SELECT distinct acc.user_id,
            acc.s3_access_name,
            acc.s3_access_key,
            acc.s3_secret_key,
            endp.s3_default_region,
            endp.url,
            endp.name,
            endp.trust_ca_bundle
           FROM ( SELECT ug.user_id,
                    s3a.name AS s3_access_name,
                    s3a.s3_access_key,
                    s3a.s3_secret_key,
                    s3a.s3_endpoint_id
                   FROM s3_assoc_user_group ug
                     JOIN s3_group gr ON gr.id = ug.s3_group_id
                     JOIN s3_assoc_group_s3access gs3a ON gr.id = gs3a.s3_group_id
                     JOIN s3_access s3a ON gs3a.s3_access_id = s3a.id
                UNION
                 SELECT us3a.user_id,
                    s3a.name AS s3_access_name,
                    s3a.s3_access_key,
                    s3a.s3_secret_key,
                    s3a.s3_endpoint_id
                   FROM s3_assoc_user_s3access us3a
                     JOIN s3_access s3a ON us3a.s3_access_id = s3a.id) acc
             JOIN s3_endpoint endp ON endp.id = acc.s3_endpoint_id;"""
        self.db.engine.execute(sql)

    def setDefaultEndpoints(self):
        amount = self.session.query(S3Endpoint).count()
        enpoints = {
            "aws_us_east_1": {
                "url": "s3.us-east-1.amazonaws.com",
                "region": "us-east-1"
            },
            "aws_us_west_1": {
                "url": "s3.us-west-1.amazonaws.com",
                "region": "us-west-1"
            },
            "aws_af_south_1": {
                "url": "s3.af-south-1.amazonaws.com",
                "region": "af-south-1"
            },
            "aws_ap_southeast_1": {
                "url": "s3.ap-southeast-1.amazonaws.com",
                "region": "ap-southeast-1"
            },
            "aws_ap_south_1": {
                "url": "s3.ap-south-1.amazonaws.com",
                "region": "ap-south-1"
            },
            "aws_eu_central_1": {
                "url": "s3.eu-central-1.amazonaws.com",
                "region": "eu-central-1"
            },
            "aws_eu_west_1": {
                "url": "s3.eu-west-1.amazonaws.com",
                "region": "eu-west-1"
            },
            "aws_sa_east_1": {
                "url": "s3.sa-east-1.amazonaws.com",
                "region": "sa-east-1"
            },
            "aws_me_south_1": {
                "url": "s3.me-south-1.amazonaws.com",
                "region": "me-south-1"
            },
            "aws_region": {
                "url": "s3.region.amazonaws.com",
                "region": "region"
            },
            "ionos_cloud_eu_central_1": {
                "url": "S3-eu-central-1.ionoscloud.com",
                "region": "eu-central-1"
            },
            "ionos_cloud_eu_central_2": {
                "url": "S3-eu-central-2.ionoscloud.com",
                "region": "eu-central-2"
            }
        }
        if amount < len(enpoints):
            for name, param in enpoints.items():
                endpoint = S3Endpoint(name=name, url=param['url'], s3_default_region=param['region'],
                                      trust_ca_bundle="")
                self.session.add(endpoint)
            self.session.commit()

    def setTestData(self, appbuilder):
        amount = self.session.query(User).count()
        accesses = {
            "S3Cloud1": {
                "accesskey": "00e11fd15cbc81b24ac1",
                "secretKey": "NBP7v6E7QYPRrpDlVYmg7p+uZeeAoCLK4pfGOJHc",
                "endpoint": "ionos_cloud_eu_central_2"
            },
            "S3Cloud2": {
                "accesskey": "00e11fd15cbc81b24ac1",
                "secretKey": "NBP7v6E7QYPRrpDlVYmg7p+uZeeAoCLK4pfGOJHc",
                "endpoint": "ionos_cloud_eu_central_2"
            },
            "S3Cloud3": {
                "accesskey": "00e11fd15cbc81b24ac1",
                "secretKey": "NBP7v6E7QYPRrpDlVYmg7p+uZeeAoCLK4pfGOJHc",
                "endpoint": "ionos_cloud_eu_central_1"
            }
        }
        sets = {
            "user1": {
                "Group1": "S3Cloud1"
            },
            "user2": {
                "Group1": "S3Cloud2"
            },
            "user3": {
                "Group1": "S3Cloud2"
            }
        }
        role_admin = appbuilder.sm.find_role(
            appbuilder.sm.auth_role_admin
        )
        for username, groupdict in sets.items():
            user = appbuilder.sm.add_user(
                username=username,
                first_name=username,
                last_name=username,
                email=username + "n@example.com",
                role=role_admin,
                password=username
            )
            for groupname in list(groupdict.keys()):
                group = self.session.query(S3Group).filter_by(name=groupname).first()
                if group == None:
                    group = S3Group(name=groupname)
                access = self.session.query(S3Access).filter_by(name=groupdict[groupname]).first()
                if access == None:
                    accessdict = accesses[groupdict[groupname]]
                    endpointid = self.session.query(S3Endpoint.id).filter_by(
                        name=accessdict['endpoint']).scalar_subquery()
                    access = S3Access(name=groupdict[groupname], s3_access_key=accessdict['accesskey'],
                                      s3_secret_key=accessdict['secretKey'], s3_endpoint_id=endpointid)
                group.s3_accesses.append(access)
                user.groups.append(group)
            self.session.add(user)
        self.session.commit()



from flask_appbuilder.security.sqla.manager import SecurityManager
from .s3_sec_models import S3User
from .s3_sec_views import S3UserDBModelView


class S3SecurityManager(SecurityManager):
    user_model = S3User
    userdbmodelview = S3UserDBModelView



import flask
import flask_login
from flask import redirect, url_for, session, current_app, Blueprint
from flask_login import current_user
from sqlalchemy import MetaData, create_engine

base = Blueprint("base", __name__)


@base.route('/favicon.ico')
def favicon():
    return redirect(url_for('static', filename='img/favicon.ico'), 308)


@base.route('/logout/', methods=['GET'])
def logout():
    if current_user.is_authenticated:
        current_app.logger.info(current_user.username + " logged out.")
        session[current_user.id] = None
        flask_login.logout_user()
    return flask.redirect("/login/")

@base.route('/s3app/liveness', methods=['GET'])
def liveness():
    return '', 200

@base.route('/s3app/readiness', methods=['GET'])
def readiness():
    meta = MetaData(create_engine(current_app.config['SQLALCHEMY_DATABASE_URI']))
    meta.reflect(views=True)
    if 's3_user_access' in meta.tables:
        return '', 200
    return '', 503



import os
from urllib.parse import urlsplit

import botocore
from botocore.config import Config
import boto3, json, tempfile
from types import SimpleNamespace

from flask import current_app

class S3BucketConfig:
    name = ""
    creationDate = ""
    region = ""

    def __init__(self, name, creationdate, region):
        self.name = name
        self.creationDate = creationdate
        self.region = region


class S3AccessConfig:
    s3AccessName = ""
    s3EndpointName = ""
    s3Endpoint = ""
    s3DefaultEndpointUrl = ""
    s3DefaultRegion = ""
    s3AccessKey = ""
    s3SecretKey = ""
    s3TrustCaBundle = ""
    s3Buckets = {}
    isAWS = False

    def fromJson(self, jsonStr):
        simpleNamespaceObject = json.loads(jsonStr, object_hook=lambda d: SimpleNamespace(**d))
        simpleNamespaceObject.s3Buckets = simpleNamespaceObject.s3Buckets

    def setCaBundleFile(self):
        tmpCaBundle, filename = tempfile.mkstemp()
        os.write(tmpCaBundle, bytes(self.s3TrustCaBundle, 'utf-8'))
        os.close(tmpCaBundle)
        return filename

class Filter:
    s3FilterFolder: bool
    s3FilterFile: bool

    def __init__(self):
        self.s3FilterFolder = True
        self.s3FilterFile = True


class S3WebData:
    accessList = []
    currentBucketList = []
    currentAccessName = ""
    currentBucketName = ""
    currentPrefix = ""
    currentSearchPrefix = ""
    contents = []
    commonPrefixes = []
    isTruncated: bool
    pageAmount: int
    entriesAmount: int
    currentPageNumber: int = 1
    page = {}
    name: str
    maxKeys: int = 20
    keyCount: int

    def setS3ListObjects(self, s3ListObjects):
        current_app.logger.debug("Set data for web frontend.")
        self.contents = []
        if 'Contents' in s3ListObjects:
            contentList = []
            for keyData in s3ListObjects['Contents']:
                if not (keyData['Key'][-1] == '/' and keyData['Size'] == 0):
                    contentList.append(keyData)
            self.contents = json.loads(json.dumps(contentList, default=str))
        self.commonPrefixes = []
        if 'CommonPrefixes' in s3ListObjects:
            for comPrefData in s3ListObjects['CommonPrefixes']:
                self.commonPrefixes.append(comPrefData['Prefix'])
        self.isTruncated = s3ListObjects['IsTruncated']
        self.name = s3ListObjects['Name']
        self.maxKeys = s3ListObjects['MaxKeys']
        self.keyCount = s3ListObjects['KeyCount']
        current_app.logger.debug("Web frontend data is: %s", json.dumps(self.__dict__))

    def setAccessList(self, s3AccessConfigs):
        current_app.logger.debug("Set list of accesses of th user.")
        self.accessList = []
        for accessname in s3AccessConfigs.keys():
            accessItem = {}
            accessItem['id'] = accessname
            accessItem['text'] = accessname
            self.accessList.append(accessItem)
        current_app.logger.debug("Access list is: %s", json.dumps(self.accessList))
        return len(self.accessList)

    def setBucketList(self, s3Buckets):
        current_app.logger.debug("Set list of buckets of the user and access.")
        self.currentBucketList = []
        for bucketname, bucketconfig in s3Buckets.items():
            bucketItem = {}
            bucketItem['id'] = bucketname
            bucketItem['text'] = bucketname
            bucketItem['creationdate'] = bucketconfig.creationDate
            bucketItem['region'] = bucketconfig.region
            self.currentBucketList.append(bucketItem)
        current_app.logger.debug("Current bucket list is: %s", json.dumps(self.currentBucketList))
        return len(self.currentBucketList)


class S3:
    s3AccessConfigs = {}  # name: S3AccessConfig
    s3WebData = {}
    access = None
    pages = {}

    def __init__(self):
        self.s3WebData = S3WebData()


    def initAccesses(self):
        self.s3WebData.currentAccessName = list(self.s3AccessConfigs.keys())[0]
        self.setInitialBucketname()
        current_app.logger.debug("Current access name is %s", self.s3WebData.currentAccessName)

    def setInitialBucketname(self):
        buckKeys = self.s3AccessConfigs[self.s3WebData.currentAccessName].s3Buckets.keys()
        if len(buckKeys) > 0:
            self.s3WebData.currentBucketName = list(buckKeys)[0]
        else:
            self.s3WebData.currentBucketName = None
        current_app.logger.debug("Current bucket name is %s", self.s3WebData.currentBucketName)

    def getBoto3Client(self, region_name=None):
        verifyPath=None
        s3AccessConfig = self.s3AccessConfigs[self.s3WebData.currentAccessName]
        if s3AccessConfig.s3TrustCaBundle != "":
            verifyPath = s3AccessConfig.setCaBundleFile()
        client = {}
        if region_name is None:
            client = boto3.client(
                's3',
                aws_access_key_id=self.s3AccessConfigs[self.s3WebData.currentAccessName].s3AccessKey,
                aws_secret_access_key=self.s3AccessConfigs[self.s3WebData.currentAccessName].s3SecretKey,
                verify=verifyPath
            )
        else:
            client = boto3.client(
                's3',
                aws_access_key_id=self.s3AccessConfigs[self.s3WebData.currentAccessName].s3AccessKey,
                aws_secret_access_key=self.s3AccessConfigs[self.s3WebData.currentAccessName].s3SecretKey,
                region_name=region_name,
                verify=verifyPath
            )
        if not self.s3AccessConfigs[self.s3WebData.currentAccessName].isAWS:
            client.endpointUrls = self.s3AccessConfigs[self.s3WebData.currentAccessName].s3DefaultEndpointUrl
        current_app.logger.debug("Boto Client defined.")
        return client

    def getBuckets(self):
        s3client = self.getBoto3Client()
        return s3client.list_buckets()

    def setBucketItems(self, prefix):
        current_app.logger.debug("Begin to list objects in current bucket.")
        if self.s3WebData.currentPrefix != prefix:
            self.setPages(prefix)
        self.s3WebData.currentPrefix = prefix
        s3client = self.getBoto3Client()
        try:
            if self.s3WebData.currentPageNumber > 1:
                s3listobjects = s3client.list_objects_v2(Bucket=self.s3WebData.currentBucketName, Prefix=prefix,
                                                         Delimiter="/", MaxKeys=self.s3WebData.maxKeys,
                                                         StartAfter=self.pages[self.s3WebData.currentPageNumber])
            else:
                s3listobjects = s3client.list_objects_v2(Bucket=self.s3WebData.currentBucketName, Prefix=prefix,
                                                         Delimiter="/", MaxKeys=self.s3WebData.maxKeys)
        except botocore.exceptions.ClientError as err:
            if err.response['Error']['Code'] == 'InternalError':  # Generic error
                # We grab the message, request ID, and HTTP code to give to customer support
                print('Error Message: {}'.format(err.response['Error']['Message']))
                print('Request ID: {}'.format(err.response['ResponseMetadata']['RequestId']))
                print('Http code: {}'.format(err.response['ResponseMetadata']['HTTPStatusCode']))
            else:
                raise err
        current_app.logger.debug("Objects in listing to visualize: %s", s3listobjects)
        self.s3WebData.setS3ListObjects(s3listobjects)

    def getPresignedUrl(self, key):
        current_app.logger.debug("Define presigned URL for key: %s and bucket: %s", key, self.s3WebData.currentBucketName)
        s3client = self.getBoto3Client(self.s3AccessConfigs[self.s3WebData.currentAccessName].s3Buckets[self.s3WebData.currentBucketName].region)
        preSignedURL = s3client.generate_presigned_url('get_object', Params={'Bucket': self.s3WebData.currentBucketName,
                                                                              'Key': key}, ExpiresIn=3600)
        return preSignedURL

    def setPages(self, prefix):
        current_app.logger.debug("Define pages for prefix: %s", prefix)
        self.s3WebData.currentPageNumber = 1
        self.s3WebData.pageAmount = 0
        self.s3WebData.entriesAmount = 0
        s3client = self.getBoto3Client()
        paginator = s3client.get_paginator('list_objects_v2')
        operation_parameters = {"Bucket": self.s3WebData.currentBucketName, "Prefix": prefix,
                                "Delimiter": "/", "MaxKeys": self.s3WebData.maxKeys}
        current_app.logger.debug("Pagination parameter are: %s", json.dumps(operation_parameters))
        page_iterator = paginator.paginate(**operation_parameters)
        for page in page_iterator:
            self.s3WebData.pageAmount = self.s3WebData.pageAmount + 1
            self.s3WebData.entriesAmount = self.s3WebData.entriesAmount + page['KeyCount']
            if "ContinuationToken" in page:
                self.pages[self.s3WebData.pageAmount] = page["ContinuationToken"]
            else:
                self.pages[self.s3WebData.pageAmount] = ""
        current_app.logger.debug("Page setting is: %s", json.dumps(self.pages))
        if self.s3WebData.pageAmount == 1 and prefix != "" and self.s3WebData.currentSearchPrefix == "":  # bug in IONOS Cloud S3?
            self.s3WebData.entriesAmount = self.s3WebData.entriesAmount - 1





import sys
import traceback

import flask_login
from flask import render_template, Blueprint, current_app, session, redirect, flash
from flask_login import current_user


def logout_user(e):
    flash("An internal error occurred. You are automatically logged out.", 'warning')
    current_app.logger.error("You will be logged due to an error: {error}".format(error=e))
    traceback.print_exc(file=sys.stdout)
    if current_user.is_authenticated:
        current_app.logger.info(current_user.username + " logged out.")
        session[current_user.id] = None
        flask_login.logout_user()
    return redirect("/login/")

def handle_500(e):
    flash("An internal error occurred. You are automatically logged out.", 'warning')
    if current_user.is_authenticated:
        current_app.logger.info(current_user.username + " logged out.")
        session[current_user.id] = None
        flask_login.logout_user()
    return redirect("/login/")


import argparse
import os, logging
import sys
from os.path import join, dirname
from pathlib import Path
from waitress import serve
import yaml
from dotenv import dotenv_values
import flask_login
from flask import Flask, flash
from flask_appbuilder import SQLA, AppBuilder
from flask_session import Session
from flask_appbuilder.menu import Menu
from werkzeug.exceptions import InternalServerError
from werkzeug.middleware.proxy_fix import ProxyFix
from sqlalchemy import MetaData, create_engine

from s3app.s3_server import S3ServerAroParser
from s3app.s3_sec import S3SecurityManager
from s3app.s3_base import base
from s3app.s3_error import logout_user, handle_500
from s3app.s3_sec_models import Sqls
from s3app.s3_sec_views import S3AccessModelView, S3GroupModelView, S3EndpointModelView
from s3app.s3_content_view import S3View, S3IndexView
from s3app.s3_sec_signals import userloggedin, userloggedout
from s3app.s3_content_rest import Access, Bucket, Page, MaxKeys
from flask_restful import Api

basedir = os.path.abspath(os.path.dirname(__file__))
# init Flask
app = Flask(__name__)
app.wsgi_app = ProxyFix(app.wsgi_app, x_proto=1)

app.config.from_pyfile(basedir + "/config.py")
log_level = logging.INFO
if os.getenv("S3APP_CONF_FILE") is not None:
    dotenv_path = os.getenv("S3APP_CONF_FILE")
    if not os.path.isfile(dotenv_path):
        print("ERROR: Cannot find .env File in {path}. Cannot start.".format(path=dotenv_path))
        sys.exit(1)
    app.config.update(dotenv_values(dotenv_path))
    app.config['SECRET_KEY'] = app.config['S3APP_SECRET_KEY'] if 'S3APP_SECRET_KEY' in app.config else app.config[
        'SECRET_KEY']
    log_level = logging.DEBUG if app.config['S3APP_LOG_LEVEL'].upper() == "DEBUG" else log_level
    log_level = logging.WARNING if app.config['S3APP_LOG_LEVEL'].upper() == "WARNING" else log_level
    log_level = logging.FATAL if app.config['S3APP_LOG_LEVEL'].upper() == "FATAL" else log_level
    log_level = logging.INFO if app.config['S3APP_LOG_LEVEL'].upper() == "INFO" else log_level
    log_level = logging.ERROR if app.config['S3APP_LOG_LEVEL'].upper() == "ERROR" else log_level

logging.getLogger().setLevel(log_level)
log_format = "%(asctime)s:%(levelname)s:%(name)s:%(message)s"

if 'S3APP_LOG_FILE' in app.config:
    try:
        with open(app.config['S3APP_LOG_FILE'], 'w') as f:
            f.write('S3App Logfile')
        logging.basicConfig(filename=app.config['S3APP_LOG_FILE'], format=log_format)
    except FileNotFoundError as e:
        print("Cannot create log file. Stop processing. Error: {error}".format(error=e))
        sys.exit(1)
else:
    logging.basicConfig(format=log_format)

app.logger.info("The APP is running in {LOGLEVEL} mode.".format(LOGLEVEL=logging.getLevelName(log_level)))

if "S3APP_APP_NAME" in app.config:
    app.config["APP_NAME"] = app.config["S3APP_APP_NAME"]
app.logger.info("Starting S3App with name {APP_NAME}.".format(APP_NAME=app.config["APP_NAME"]))

if "S3APP_APP_ICON" in app.config:
    app.config["APP_ICON"] = app.config["S3APP_APP_ICON"]
app.logger.info("Adding {APP_ICON} as App icon.".format(APP_ICON=app.config["APP_ICON"]))

# Init Database
server_session = Session(app)
server_session.app.session_interface.db.create_all()
db = SQLA(app)

# Register Flask Blueprints
app.register_blueprint(base)

# Register Flask error handler
app.register_error_handler(InternalServerError, handle_500)
app.register_error_handler(401, logout_user)

# Init Appbuilder Authentication and Authorization
appbuilder = AppBuilder(app, db.session, base_template='appbuilder/baselayout.html',
                        security_manager_class=S3SecurityManager, indexview=S3View, menu=Menu(reverse=False))
appbuilder.add_view(S3AccessModelView, "List S3Access", category="Security")
appbuilder.add_view(S3GroupModelView, "List Groups", category="Security")
appbuilder.add_view(S3EndpointModelView, "List S3Endpoints", category="Security")
appbuilder.add_view_no_menu(S3View())
appbuilder.add_view_no_menu(S3IndexView())

# init Flask RESTApi
api = Api(app)
api.add_resource(Access, '/s3config/access/<string:accessname>')
api.add_resource(Bucket, '/s3config/access/<string:accessname2>/bucket/<string:bucketname>')
api.add_resource(MaxKeys, '/s3config/maxkeys/<string:maxkeys>')
api.add_resource(Page, '/s3config/page/<string:page>')

admin = appbuilder.sm.find_user(username="admin")

# Setup default Database
if admin == None:
    role_admin = appbuilder.sm.find_role(
        appbuilder.sm.auth_role_admin
    )
    admin = appbuilder.sm.add_user(
        username="admin",
        first_name="admin",
        last_name="admin",
        email="admin@example.com",
        role=role_admin,
        password="admin"
    )
    db.session.add(admin)
    db.session.commit()

# Init signal triggers
flask_login.user_logged_in.connect(userloggedin)
flask_login.user_logged_out.connect(userloggedout)

# Init access views

sqls = Sqls(db)

sqls.setDefaultEndpoints()
if os.environ.get('WITH_TESTDATA') == "true":
    sqls.setTestData(appbuilder)

meta = MetaData(create_engine(app.config['SQLALCHEMY_DATABASE_URI']))
meta.reflect(views=True)
if not 's3_user_access' in meta.tables:
    sqls.setUserAccessView()


def run():
    argParser = S3ServerAroParser()
    args = argParser.parser.parse_args()
    serve(app, **argParser.getKw(args))



from flask import session, current_app
from flask_login import current_user
from flask_restful import Resource, abort


class Access(Resource):
    def post(self, accessname):
        if current_user.is_authenticated:
            current_app.logger.debug("Change access to %s for user %s", accessname, current_user.username)
            s3 = session[current_user.id]
            s3.s3WebData.currentAccessName = accessname
            s3.s3WebData.currentSearchPrefix = ""
            s3.s3WebData.setBucketList(s3.s3AccessConfigs[s3.s3WebData.currentAccessName].s3Buckets)
            s3.setInitialBucketname()
            if s3.s3WebData.currentBucketName is not None:
                s3.setPages("")
            session[current_user.id] = s3
            return "", 200
        else:
            current_app.logger.info("Anonymous access to POST Access")
            abort(401)


class Bucket(Resource):
    def post(self, accessname2, bucketname):
        if current_user.is_authenticated:
            current_app.logger.debug("Change bucket to %s for user %s", bucketname, current_user.username)
            s3 = session[current_user.id]
            s3.s3WebData.currentAccessName = accessname2
            s3.s3WebData.currentBucketName = bucketname
            s3.s3WebData.currentSearchPrefix = ""
            s3.setPages("")
            session[current_user.id] = s3
            return "", 200
        else:
            current_app.logger.info("Anonymous access to POST Bucket")
            abort(401)


class MaxKeys(Resource):
    def post(self, maxkeys):
        if current_user.is_authenticated:
            current_app.logger.debug("Change MaxKeys to %s for user %s", maxkeys, current_user.username)
            s3 = session[current_user.id]
            s3.s3WebData.maxKeys = int(maxkeys)
            session[current_user.id] = s3
            s3.setPages("")
            return "", 200
        else:
            current_app.logger.info("Anonymous access to POST MaxKeys")
            abort(401)


class Page(Resource):
    def post(self, page):
        if current_user.is_authenticated:
            current_app.logger.debug("Change Page to %s for user %s", page, current_user.username)
            s3 = session[current_user.id]
            s3.s3WebData.currentPageNumber = int(page)
            session[current_user.id] = s3
            return "", 200
        else:
            current_app.logger.info("Anonymous access to POST Page")
            return '', 401




import boto3
from flask import session, flash, current_app
from sqlalchemy import MetaData, create_engine
from .s3_content_class import S3, S3AccessConfig, S3BucketConfig

def userloggedin(app, user):
    current_app.logger.info(user.username + " logged in.")
    meta = MetaData(create_engine(app.config['SQLALCHEMY_DATABASE_URI']))
    meta.reflect(views=True)
    accmodel = meta.tables['s3_user_access']
    s3AccessConfigs = {}
    s3 = S3()
    for result in app.appbuilder.session.query(accmodel).filter_by(user_id=user.id).all():
        s3AccessConfig = S3AccessConfig()
        s3AccessConfig.s3AccessName = result['s3_access_name']
        s3AccessConfig.s3AccessKey = result['s3_access_key']
        s3AccessConfig.s3SecretKey = result['s3_secret_key']
        s3AccessConfig.s3DefaultEndpointUrl = "https://" + result['url']
        s3AccessConfig.s3DefaultRegion = result['s3_default_region']
        s3AccessConfig.s3Endpoint = result['url']
        s3AccessConfig.s3EndpointName = result['name']
        s3AccessConfig.s3TrustCaBundle = result['trust_ca_bundle']
        verifyPath = None
        if result['url'].endswith("amazonaws.com"):
            s3AccessConfig.isAWS = True
        if s3AccessConfig.s3TrustCaBundle != "":
            verifyPath = s3AccessConfig.setCaBundleFile()
        s3conn = boto3.client('s3',
                              aws_access_key_id=s3AccessConfig.s3AccessKey,
                              aws_secret_access_key=s3AccessConfig.s3SecretKey,
                              endpoint_url=s3AccessConfig.s3DefaultEndpointUrl,
                              region_name=s3AccessConfig.s3DefaultRegion,
                              verify=verifyPath
                              )
        response = s3conn.list_buckets()
        buckets = {}
        for bucket in response['Buckets']:
            bucketLocation = s3conn.get_bucket_location(Bucket=bucket['Name'])
            bucketLocName = ""
            if bucketLocation['LocationConstraint'] is not None:
                bucketLocName = bucketLocation['LocationConstraint']
            buckets[bucket['Name']] = S3BucketConfig(bucket['Name'], bucket['CreationDate'].utcnow().isoformat(),
                                                     bucketLocName)
        s3AccessConfig.s3Buckets = buckets
        s3AccessConfigs[s3AccessConfig.s3AccessName] = s3AccessConfig
    if len(s3AccessConfigs.keys()) > 0:
        s3.s3AccessConfigs = s3AccessConfigs
        s3.initAccesses()
        s3.s3WebData.setBucketList(s3.s3AccessConfigs[s3.s3WebData.currentAccessName].s3Buckets)
        s3.s3WebData.setAccessList(s3.s3AccessConfigs)
        if s3.s3WebData.currentBucketName != "":
            s3.setPages("")
    else:
        current_app.logger.info(user.username + " das no access configured.")
    session[user.id] = s3


def userloggedout(app, user):
    flash("You are logged out", 'info')
